/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
File    : RTOSInit_RX66T.c
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ  (160000000u / 2u / CMT0_TIMER_PRESCALE)
#define OS_TICK_FREQ   (1000u)
#define OS_INT_FREQ    (OS_TICK_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKRX.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

/*********************************************************************
*
*       Device specific SFRs
*/
//
// Register Write Protection Function
//
#define PRCR                  (*(volatile EVENACCESS OS_U16*) (0x000803FEu))

//
// Interrupt controller
//
#define IRR_BASE_ADDR        (0x00087000u)
#define ISEL_BASE_ADDR       (0x00087100u)

#define IER03                (*(volatile EVENACCESS OS_U8*)  (0x00087203u))
#define IER1A                (*(volatile EVENACCESS OS_U8*)  (0x0008721Au))
#define IER1B                (*(volatile EVENACCESS OS_U8*)  (0x0008721Bu))

#define IPR04                (*(volatile EVENACCESS OS_U8*)  (0x00087304u))
#define IPR80                (*(volatile EVENACCESS OS_U8*)  (0x00087380u))

//
// SYSTEM module
//
#define SYSTEM_BASE_ADDR     (0x00080000u)
#define SYSCR0               (*(volatile EVENACCESS OS_U16*) (SYSTEM_BASE_ADDR + 0x0006u))
#define MSTPCRA              (*(volatile EVENACCESS OS_U32*) (SYSTEM_BASE_ADDR + 0x0010u))
#define SCKCR                (*(volatile EVENACCESS OS_U32*) (SYSTEM_BASE_ADDR + 0x0020u))
#define SCKCR2               (*(volatile EVENACCESS OS_U16*) (SYSTEM_BASE_ADDR + 0x0024u))
#define SCKCR3               (*(volatile EVENACCESS OS_U16*) (SYSTEM_BASE_ADDR + 0x0026u))
#define PLLCR                (*(volatile EVENACCESS OS_U16*) (SYSTEM_BASE_ADDR + 0x0028u))
#define PLLCR2               (*(volatile EVENACCESS OS_U8 *) (SYSTEM_BASE_ADDR + 0x002Au))
#define MOSCCR               (*(volatile EVENACCESS OS_U8 *) (SYSTEM_BASE_ADDR + 0x0032u))
#define OSCOVFSR             (*(volatile EVENACCESS OS_U8 *) (SYSTEM_BASE_ADDR + 0x003Cu))
#define MOSCWTCR             (*(volatile EVENACCESS OS_U8 *) (SYSTEM_BASE_ADDR + 0x00A2u))
#define MEMWAIT              (*(volatile EVENACCESS OS_U8 *) (SYSTEM_BASE_ADDR + 0x101Cu))
#define MOFCR                (*(volatile EVENACCESS OS_U8 *) (SYSTEM_BASE_ADDR + 0xC293u))

//
// Assign timer used for OS
//
#define CMT0_BASE_ADDR       (0x00088000u)
#define CMT0_MSTPCR_BIT      (15u)
#define CMT0_VECT            28u
#define CMT0_IPR             (IPR04)
#define CMT0_IER             (IER03)
#define CMT0_IER_BIT         (4u)
#define CMT0_INT_PRIO        (0x01u)  // Lowest priority
#define CMT0_TIMER_PRESCALE  (8u)
#define CMT0_CMSTR0          (*(volatile EVENACCESS OS_U16*) (CMT0_BASE_ADDR + 0x00u))
#define CMT0_CMCR            (*(volatile EVENACCESS OS_U16*) (CMT0_BASE_ADDR + 0x02u))
#define CMT0_CMCNT           (*(volatile EVENACCESS OS_U16*) (CMT0_BASE_ADDR + 0x04u))
#define CMT0_CMCOR           (*(volatile EVENACCESS OS_U16*) (CMT0_BASE_ADDR + 0x06u))

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USE_PARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USE_PARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return CMT0_CMCNT;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return *(OS_U8*)(IRR_BASE_ADDR + CMT0_VECT);
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       SysTick_Handler()
*
*  Function description
*    This is the hardware timer exception handler.
*/
#ifdef __ICCRX__
  #pragma vector = (CMT0_VECT)
  static __interrupt void SysTick_Handler (void) {
#endif
#ifdef __RX
  #pragma interrupt (SysTick_Handler(vect=CMT0_VECT))
  static void SysTick_Handler (void) {
#endif
#ifdef __GNUC__
  void __attribute__ ((interrupt)) SysTick_Handler (void) {
#endif
#if (OS_SUPPORT_TRACE_API != 0)
  SEGGER_SYSVIEW_TickCnt++;
#endif
  OS_INT_EnterNestable();
  OS_TICK_Handle();
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKRX_Process();
#endif
  OS_INT_LeaveNestable();
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
void OS_InitHW(void) {
  OS_INT_IncDI();
  //
  // Initialize timer for embOS, assuming PLL is already initialized
  //
  PRCR         = 0xA50Bu;                                  // Disable register protection
  MSTPCRA     &= ~(1u << CMT0_MSTPCR_BIT);                 // Activate CMT unit
  CMT0_CMSTR0 &= ~1u;                                      // Stop counter
  CMT0_CMCR    = (1u << 6u);                               // Compare match interrupt enable, PCLK/8
  CMT0_CMCOR   = (OS_U16)(OS_TIMER_FREQ / OS_TICK_FREQ);   // Count value for 1 msec resolution
  //
  // Enable CMT0 interrupt in interrupt controller
  //
  CMT0_IPR     = CMT0_INT_PRIO;                            // Set interrupt priority, lowest priority
  CMT0_IER    |= (1u << CMT0_IER_BIT);                     // Enable CMT interrupt
  *(OS_U8*)(IRR_BASE_ADDR + CMT0_VECT)  = 0u;              // Clear interrupt request
  *(OS_U8*)(ISEL_BASE_ADDR + CMT0_VECT) = 0u;              // Select interrupt source CPU
  CMT0_CMSTR0 |= 1u;                                       // Start counter
  PRCR         = 0xA500u;                                  // Enable register protection
  //
  // Inform embOS about the timer settings
  //
  {
    OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
    OS_TIME_ConfigSysTimer(&SysTimerConfig);
  }
  //
  // Configure and initialize SEGGER SystemView
  //
#if (OS_SUPPORT_TRACE_API != 0)
  SEGGER_SYSVIEW_Conf();
#endif
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKRX_SetpfOnRx(OS_COM_OnRx);
  JLINKRX_SetpfOnTx(OS_COM_OnTx);
  JLINKRX_SetpfGetNextChar(OS_COM_GetNextChar);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {  // Idle loop: No task is ready to execute
  while (1) {         // Nothing to do ... wait for interrupt
#if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
  #ifdef __ICCRX__
    __wait_for_interrupt();
  #endif
  #ifdef __RX
    wait();
  #endif
  #ifdef __GNUC__
    __builtin_rx_wait();
  #endif
#endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKRX_SendChar(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USE_PARA(c);          // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       HardwareSetup()
*
*  Function description
*    Initialize clock generation and PLL
*
*    Has to be modified if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before HardwareSetup() is called,
*    no access or call which relies on any variable can be used here
*    and no variables can be set to a specific value as long as those variables
*    are not declared as "no init".
*
*    This function is called from startup code.
*/
void HardwareSetup(void) {
  //
  // Initialize PLL, assuming external crystal of 8.0 MHz
  //
  PRCR     = 0xA50Bu;                    // Disable register protection
  MOFCR    = 0u
           | (3u << 4u)                  // MODRV : Driving 8MHz
           | (0u << 6u)                  // MOSEL : Resonator
           ;
  MOSCWTCR = 0u;                         // MSTS    : Zero waiting time for external clock signal
  MOSCCR   = 0u;                         // MOSTP   : Enable main clock oscillator
  MEMWAIT  = 1u;                         // MEMWAIT : Set to 1 wait cycle (for ICLK > 120MHz)
  PLLCR    = 0u
           | ( 0u << 0u)                 // PLIDIV    : Set divisor to 1 (8MHz / 1 = 8MHz)
           | ( 0u << 2u)                 // reserved
           | ( 0u << 4u)                 // PLLSRCSEL : main clock oscillator
           | ( 0u << 5u)                 // reserved
           | (39u << 8u)                 // STC       : Set multiplication factor to 20 (8MHz * 20 = 160MHz)
           ;
  PLLCR2   = 0u;                         // Enable PLL
  while ((OSCOVFSR & (1u << 2u)) == 0);  // Wait for the PLL to become stable
  SCKCR    = 0u
           | (0u <<  0u)                 // PCKD     : *1/1
           | (0u <<  4u)                 // PCKC     : *1/1
           | (1u <<  8u)                 // PCKB     : *1/2
           | (0u << 12u)                 // PCKA     : *1/1
           | (0u << 16u)                 // BCK      : *1/1
           | (0u << 20u)                 // b22..b20 : reserved
           | (0u << 23u)                 // PSTOP1   : BCLK pin output enabled
           | (0u << 24u)                 // ICK      : *1/1
           | (0u << 28u)                 // FCK      : *1/1
           ;
  SCKCR2   = 1u                          // reserved b0 shall be written 1
           | (1u << 4u)                  // UCK : *1/2
           ;
  SCKCR3   = (4u << 8u);                 // CLKSEL : Select PLL
  PRCR     = 0xA500u;
}

#ifdef __ICCRX__
/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL
*
*    Has to be modified, if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
int __low_level_init(void);  // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
int __low_level_init(void) {
  HardwareSetup();
  return 1;  // Always return 1 to enable segment initilization !
}
#endif

/*************************** End of file ****************************/
