/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : UART implementation

Additional information:

  Device : RX72T
  Board  : Renesas RSKRX72T

  Unit | UART   | Board connector
  ===============================
  0    | SCI11  | G1CUSB0 (SCI11 Tx PB5 Pin 41)
       |        |         (SCI11 Rx PB6 Pin 40)

*/

#include "BSP_UART.h"
#include "RTOS.h"  // For OS_INT_Enter()/OS_INT_Leave(). Remove this line and OS_INT_* functions if not using embOS.

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/
#define BSP_UART_CLOCK     (192000000u / 2u)  // SCI11 derives from PCLKA
#define BSP_UART_BAUDRATE  (38400u)

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/
#ifndef EVENACCESS
  #if defined(__RX)
    #define EVENACCESS  __evenaccess
  #else
    #define EVENACCESS
  #endif
#endif

#define SYSTEM_BASE_ADDR          (0x00080000u)
#define SYSTEM_MSTPCRC            (*(volatile EVENACCESS unsigned int*)  (SYSTEM_BASE_ADDR + 0x018u))
#define SYSTEM_PRCR               (*(volatile EVENACCESS unsigned short*)(SYSTEM_BASE_ADDR + 0x3FEu))

#define SCI_BASE_ADDR             (0x000D0000u)  // SCI11
#define SCI_SMR                   (*(volatile EVENACCESS unsigned char*)(SCI_BASE_ADDR + 0x000u))
#define SCI_BRR                   (*(volatile EVENACCESS unsigned char*)(SCI_BASE_ADDR + 0x001u))
#define SCI_SCR                   (*(volatile EVENACCESS unsigned char*)(SCI_BASE_ADDR + 0x002u))
#define SCI_TDR                   (*(volatile EVENACCESS unsigned char*)(SCI_BASE_ADDR + 0x003u))
#define SCI_SSR                   (*(volatile EVENACCESS unsigned char*)(SCI_BASE_ADDR + 0x004u))
#define SCI_RDR                   (*(volatile EVENACCESS unsigned char*)(SCI_BASE_ADDR + 0x005u))
#define SCI_SCMR                  (*(volatile EVENACCESS unsigned char*)(SCI_BASE_ADDR + 0x006u))

#define ICU_BASE_ADDR             (0x00087000u)
#define ICU_IR114                 (*(volatile EVENACCESS unsigned char*)(ICU_BASE_ADDR + 0x072u))  // Interrupt request register
#define ICU_IR115                 (*(volatile EVENACCESS unsigned char*)(ICU_BASE_ADDR + 0x073u))  // Interrupt request register
#define ICU_IER0E                 (*(volatile EVENACCESS unsigned char*)(ICU_BASE_ADDR + 0x20Eu))  // Interrupt request enable register
#define ICU_IPR114                (*(volatile EVENACCESS unsigned char*)(ICU_BASE_ADDR + 0x372u))  // Interrupt source priority register (RXI11)
#define ICU_IPR115                (*(volatile EVENACCESS unsigned char*)(ICU_BASE_ADDR + 0x373u))  // Interrupt source priority register (TXI11)

#define MPC_PWPR                  (*(volatile EVENACCESS unsigned char*)(0x0008C11Fu))
#define MPC_PB5PFS                (*(volatile EVENACCESS unsigned char*)(0x0008C19Du))  // Pin function register
#define MPC_PB6PFS                (*(volatile EVENACCESS unsigned char*)(0x0008C19Eu))  // Pin function register
#define PORTB_PMR                 (*(volatile EVENACCESS unsigned char*)(0x0008C06Bu))  // Port mode register

#define SYSTEM_MSTPCRC_SCI11_BIT  (24u)
#define ICU_IER1D_RX_BIT          ( 2u)
#define ICU_IER1D_TX_BIT          ( 3u)
#define SCI_RX_VECT               114u
#define SCI_TX_VECT               115u
#define SCI_INT_PRIO              ( 1u)     // Lowest priority
#define SCI_ERR_PER_BIT           ( 3u)
#define SCI_ERR_FER_BIT           ( 4u)
#define SCI_ERR_ORER_BIT          ( 5u)
#define SCI_ERR_PER_MASK          (1u << SCI_ERR_PER_BIT)
#define SCI_ERR_FER_MASK          (1u << SCI_ERR_FER_BIT)
#define SCI_ERR_ORER_MASK         (1u << SCI_ERR_ORER_BIT)
#define SCI_ERR_MASK_ALL          (SCI_ERR_PER_MASK | SCI_ERR_FER_MASK | SCI_ERR_ORER_MASK)
#define SCI_BRR_BAUDDIVIDE        (((BSP_UART_CLOCK + (BSP_UART_BAUDRATE * 16u)) / (BSP_UART_BAUDRATE * 32u)) - 1u)

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
static BSP_UART_TX_CB* _pfWriteCB;
static BSP_UART_RX_CB* _pfReadCB;

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _SetBaudrate()
*
*  Function description
*    Configures the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
static void _SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  BSP_UART_USE_PARA(Unit);
  BSP_UART_USE_PARA(Baudrate);
}

/*********************************************************************
*
*       Global functions, IRQ handler
*
**********************************************************************
*/

/*********************************************************************
*
*       _ISR_Rx()
*
*  Function description
*    Interrupt handler for UART reception
*/
#if defined(__ICCRX__)
  #pragma vector = (SCI_RX_VECT)                // Rx data interrupt
  static __interrupt void _ISR_Rx(void);
  static __interrupt void _ISR_Rx(void) {
#elif defined(__RX)
  #pragma interrupt (_ISR_Rx(vect=SCI_RX_VECT)) // Rx data interrupt
  static void _ISR_Rx(void) {
#elif defined(__GNUC__)
  void __attribute__ ((interrupt)) _ISR_Rx(void) {
#endif
  unsigned char Data;

  while (SCI_SSR & (1u << 6)) {
    OS_INT_EnterNestable();    //  We will enable interrupts
    Data = SCI_RDR;
    if (_pfReadCB != NULL) {
      _pfReadCB(0, Data);      // Process actual byte
    }
    OS_INT_LeaveNestable();
  }
}

/*********************************************************************
*
*       _ISR_Tx()
*
*  Function description
*    Interrupt handler for UART transmission
*/
#if defined(__ICCRX__)
  #pragma vector = (SCI_TX_VECT)  // TX data empty interrupt
  static __interrupt void _ISR_Tx(void);
  static __interrupt void _ISR_Tx(void) {
#elif defined(__RX)
  #pragma interrupt (_ISR_Tx(vect=SCI_TX_VECT))  // TX data empty interrupt
  static void _ISR_Tx(void) {
#elif defined(__GNUC__)
  void __attribute__ ((interrupt)) _ISR_Tx(void) {
#endif
  OS_INT_EnterNestable();
  if (_pfWriteCB != NULL) {
    _pfWriteCB(0);
  }
  OS_INT_LeaveNestable();
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_UART_Init()
*
*  Function description
*    Initializes the UART peripheral.
*
*  Parameters
*    Unit       : Unit number (typically zero-based).
*    Baudrate   : Baud rate to configure [Hz].
*    NumDataBits: Number of data bits to use.
*    Parity     : One of the following values:
*                   * BSP_UART_PARITY_NONE
*                   * BSP_UART_PARITY_ODD
*                   * BSP_UART_PARITY_EVEN
*    NumStopBits: Number of stop bits to use.
*/
void BSP_UART_Init(unsigned int Unit, unsigned long Baudrate, unsigned char NumDataBits, unsigned char Parity, unsigned char NumStopBits) {
  volatile unsigned int i;

  //
  // Unused parameters. Hard coded USART settings.
  //
  BSP_UART_USE_PARA(Unit);
  BSP_UART_USE_PARA(Baudrate);
  BSP_UART_USE_PARA(NumDataBits);
  BSP_UART_USE_PARA(Parity);
  BSP_UART_USE_PARA(NumStopBits);
  //
  // Enable UART Module
  //
  SYSTEM_PRCR     = 0xA502u;
  SYSTEM_MSTPCRC &= ~(1u << SYSTEM_MSTPCRC_SCI11_BIT);
  SYSTEM_PRCR     = 0xA500u;
  //
  // Set uart port pins
  //
  MPC_PWPR    = 0x00u;   // Disable write protect
  MPC_PWPR    = 0x40u;   // Enable write access
  MPC_PB5PFS  = 0x0Bu;   // Set pin function register to uart function
  MPC_PB6PFS  = 0x0Bu;   // Set pin function register to uart function
  MPC_PWPR    = 0x80u;   // Enable write protect
  PORTB_PMR  |= 0x60u;   // Set pins to peripheral function
  //
  // Reset, disable Reception and transmission
  //
  SCI_SCR = 0x00u;
  //
  // Set transmission mode
  //
  SCI_SMR = 0x00u;
  //
  // Set transfer direction, LSB first, 8 bit data length
  //
  SCI_SCMR = 0xF2u;
  //
  // Set baudrate
  //
  SCI_BRR = (unsigned char)(SCI_BRR_BAUDDIVIDE);
  //
  // Wait for SCI settling time (1 bit cycle)
  //
  for (i = 0u; i < 3000u ; i++) {
  }
  //
  // Clear status
  //
  SCI_SSR = 0xC0u;
  //
  // Setup interrupt control register for UART
  //
  ICU_IPR114 = SCI_INT_PRIO;  // Lowest priority
  ICU_IPR115 = SCI_INT_PRIO;  // Lowest priority
  //
  // Enable UART interrupts
  //
  ICU_IER0E = ICU_IER0E
            | (1u << ICU_IER1D_TX_BIT)  // Enable Tx empty interrupt
            | (1u << ICU_IER1D_RX_BIT)  // Enable Rx interrupt
            ;
  //
  // Clear interrupt request flags
  //
  ICU_IR114 = 0u;
  ICU_IR115 = 0u;
  //
  // Finally enable Rx and Tx interrupts and enable uart Rx and Tx
  //
  SCI_SCR = SCI_SCR
          | (1u << 7u)  // Tx interrupt enable
          | (1u << 6u)  // Rx interrupt enable
          ;
  SCI_SCR = SCI_SCR
          | (1u << 5u)  // Tx enable
          | (1u << 4u)  // Rx enable
          ;
}

/*********************************************************************
*
*       BSP_UART_DeInit()
*
*  Function description
*    De-initializes the UART peripheral.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*/
void BSP_UART_DeInit(unsigned int Unit) {
  BSP_UART_USE_PARA(Unit);
}

/*********************************************************************
*
*       BSP_UART_SetBaudrate()
*
*  Function description
*    Configures/changes the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
void BSP_UART_SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  _SetBaudrate(Unit, Baudrate);
}

/*********************************************************************
*
*       BSP_UART_SetReadCallback()
*
*  Function description
*    Sets the callback to execute upon an Rx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetReadCallback(unsigned Unit, BSP_UART_RX_CB* pf) {
  BSP_UART_USE_PARA(Unit);
  _pfReadCB = pf;
}

/*********************************************************************
*
*       BSP_UART_SetWriteCallback()
*
*  Function description
*    Sets the callback to execute upon a Tx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetWriteCallback(unsigned int Unit, BSP_UART_TX_CB* pf) {
  BSP_UART_USE_PARA(Unit);
  _pfWriteCB = pf;
}

/*********************************************************************
*
*       BSP_UART_Write1()
*
*  Function description
*    Sends one byte via UART.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    Data: (First) data byte to send.
*
*  Additional information
*    The first byte of a transfer is typically sent from application
*    context. Further bytes of the transfer are then sent from the
*    Tx interrupt handler by also calling this function from interrupt
*    context.
*/
void BSP_UART_Write1(unsigned int Unit, unsigned char Data) {
  BSP_UART_USE_PARA(Unit);
  SCI_TDR = Data;
}

/*************************** End of file ****************************/
